#pragma once

// solanaceae port of sha1 fts for NGCFT1

#include <solanaceae/contact/contact_model3.hpp>
#include <solanaceae/message3/registry_message_model.hpp>
#include <solanaceae/tox_contacts/tox_contact_model2.hpp>

#include "./ngcft1.hpp"

#include "./ft1_sha1_info.hpp"

#include <entt/container/dense_map.hpp>

#include <variant>

class SHA1_NGCFT1 : public RegistryMessageModelEventI, public NGCFT1EventI {
	Contact3Registry& _cr;
	RegistryMessageModel& _rmm;
	NGCFT1& _nft;
	ToxContactModel2& _tcm;


	// limit this to each group?
	entt::dense_map<SHA1Digest, Message3Handle> _info_to_message;

	// sha1 chunk index
	// TODO: optimize lookup
	entt::dense_map<SHA1Digest, Message3Handle> _chunks;

	// group_number, peer_number, message, chunk_hash, timer
	std::deque<std::tuple<uint32_t, uint32_t, Message3Handle, SHA1Digest, float>> _queue_requested_chunk;
	//void queueUpRequestInfo(uint32_t group_number, uint32_t peer_number, const SHA1Digest& hash);
	void queueUpRequestChunk(uint32_t group_number, uint32_t peer_number, Message3Handle msg, const SHA1Digest& hash);

	struct SendingTransfer {
		struct Info {
			// copy of info data
			// too large?
			std::vector<uint8_t> info_data;
		};

		struct Chunk {
			Message3Handle msg;
			uint64_t offset_into_file;
			// or data?
			// if memmapped, this would be just a pointer
		};

		std::variant<Info, Chunk> v;

		float time_since_activity {0.f};
	};
	// key is groupid + peerid
	entt::dense_map<uint64_t, entt::dense_map<uint8_t, SendingTransfer>> _sending_transfers;

	static uint64_t combineIds(const uint32_t group_number, const uint32_t peer_number);

	public: // TODO: config
		bool _udp_only {false};

		size_t _max_concurrent_in {8};
		size_t _max_concurrent_out {4};

	public:
		SHA1_NGCFT1(
			Contact3Registry& cr,
			RegistryMessageModel& rmm,
			NGCFT1& nft,
			ToxContactModel2& tcm
		);

		void iterate(float delta);

	protected: // events
		bool onEvent(const Events::NGCFT1_recv_request&) override;
		bool onEvent(const Events::NGCFT1_recv_init&) override;
		bool onEvent(const Events::NGCFT1_recv_data&) override;
		bool onEvent(const Events::NGCFT1_send_data&) override; // const?

		bool sendFilePath(const Contact3 c, std::string_view file_name, std::string_view file_path) override;
};

